(load "ds-fp.scm")
(load "test.scm")
(load "macros.scm")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;; TESTS for distribution functionality ;;;;;;;;;;;;

;IMPORTANT
; The definitions and/or interactions window IS NOT A SERVER. Therefore:
; - it cannot export functions
; - it has no host properties
; - use printf to print in the console, sprintf to print in a server console
;
; All parameters are passed by copy, so are return values
; The return value of an *export* of a function (with export or export-as) is a stub
; Use this stub to have pass-by-reference.
;
; Every server exports the "eval" function by default:
; - implementation is (export-as (fun (thunk) (thunk)) "eval"))
; - used to send a thunk to evaluate on the server

;startup server
(start-server "localhost" "8080")
(display "----------MARK1----------")(newline)

;; checking prelude functions are available
;; a- get back the call closure
(test/pred (run 'call) closure?)
;; b- apply the true pc function (ignores its argument)
(test (run '(t #f)) #t)

;; basic tests, with localhost:8080 eval function

;creation of stubs
;; get the stub of the eval function
(test/pred (run '(lookup "localhost" "8080" "eval")) fun-stub?)

;; execute expression on remote host
;; get result back
(test (run '((lookup "localhost" "8080" "eval") (fun () (+ 3 4)))) 7)

;; checks arg count on remote call
;; uncomment this to get the exception
;; ds-fp.scm:189:8: application: bad argument count: expected 1, given 2
;(test (run '((lookup "localhost" "8080" "eval") 1 2)) 7)

;; same, using code wrapper (macro-like facility)
(test (run (on-default '(+ 3 4))) 7)

;; a- pass a thunk that defines a new function on remote host and get it back
;; expects a closure, not a fun-stub, because the created function is not remote
(test/pred (run '((lookup "localhost" "8080" "eval") (fun () (fun () 0)))) closure?)

;; b- pass a thunk with a lexically-captured variable
(test (run '((lookup "localhost" "8080" "eval") (let ((z 3)) 
                                                  (fun () z)))) 3)

;; c- same, except that the captured variable is a function
(test/pred (run '((lookup "localhost" "8080" "eval") (let ((z (fun () 5))) 
                                                       (fun () z)))) closure?)

;; d- confirm that captured variable shadows the call function of the prelude in remote host
;; (otherwise, the application '(call)' would throw an error of bad argument count 
(test (run '((lookup "localhost" "8080" "eval") (let ((call (fun () 2))) 
                                                       (fun () (call))))) 2)

;; e- get the call function from remote host
(test/pred (run (on-default 'call)) closure?)

(display "----------MARK2----------")(newline)

;; export function on remote host using a name, get back a stub
(test/pred (run (on-default '(export-as (fun (x) 340) "foo"))) fun-stub?)

;; export function on remote host (anonymous)
;; then apply it using local argument
(test (run `(,(on-default '(export (fun (x) (+ x 1)))) 10)) 11)

;; export function on remote host, by name (using a let)
;; then look it up and apply it.
(test (run `(begin 
              (let ((add1 (fun (x) (+ x 1))))
                ,(on-default '(export-as add1 "add1")))
              ((lookup "localhost" "8080" "add1") 2))) 3)
           

(display "----------MARK3----------")(newline)

;extra server used for tests below
;; starting with init code
;; defines foo constant in root environment + export foo_fun that returns foo.
(start-server-init "localhost" "9090" '(begin 
                                         (define foo 42)
                                         (export-as (fun () foo) "foo_fun")))

;; get host properties
;; a- on default server
(test (run (on-default 'current-host)) '((name . "localhost") (port . "8080")))
;; b- on other server
(test (run (on-port "9090" 'current-host)) '((name . "localhost") (port . "9090")))


;root environment lookup test
;; a- get foo directly on 9090
(test (run (on-port "9090" 'foo))
      42)
;; b- get foo_fun remotely and apply it
(test (run '((lookup "localhost" "9090" "foo_fun"))) 
      42)



(display "----------MARK4----------")(newline)

(printf "(no local effects -- check the consoles)~n")
         
;;testing printf is server-specific
;;expect nothing locally
;; printing on one server
(run (on-port "8080" '(sprintf "printing on 8080~n")))
;; and the other
(run (on-port "9090" '(sprintf "printing on 9090~n")))

;; print on 8080 from 9090 using exported function
(run `(begin 
        ,(on-port "8080" '(export-as (fun (x) (sprintf x)) "remote-print"))
        ,(on-port "9090" '((lookup "localhost" "8080" "remote-print") "printing from 9090!~n"))))


(display "----------MARK5----------")(newline)

;; ship function definition nested in let
(test (run `(let ((delta 2))
              (,(on-default '(export (fun (x) (+ x delta)))) 3))) 
      5)

;; plain local program
(test (run '(let ((count 0))
              (let ((counted (fun () (set! count (+ 1 count)) count)))
                (counted)
                (counted)))) 
      2)

;; same but remotely
(test (run (on-default '(let ((count 0))
                          (let ((counted (fun () (set! count (+ 1 count)) count)))
                            (counted)
                            (counted)))))
      2)

;; normal function in environment => passed by copy
(test (run `(let ((count 0))
              (let ((counted (fun () (set! count (+ 1 count)) count)))
                (+ ,(on-port "8080" '(counted))
                   ,(on-port "9090" '(counted))))))
      2)

;; exported function in the environment => passed by ref
(test (run (on-default 
            `(let ((count 0))
               (let ((counted (export (fun () (set! count (+ 1 count)) count))))
                 (+ ,(on-port "8080" '(counted))
                    ,(on-port "9090" '(counted)))))))
      3)



(display "---------- END ----------")(newline)
