;;;;;;;;;;;;;;;; definition of environments ;;;;;;;;;;;;;;;;

;; environments can be passed over the network and deep-copied.
(require scheme/serialize) 
(require mzlib/struct) ;; for copy-struct

(define-serializable-struct environment (ids vals env) #:mutable)
            
;; returns a copy of the given environment, in depth
;; values are treated as specified by the value transformer function
(define (env-copy e val-transf)
  (copy-struct environment e
    (environment-ids  (map (lambda (x) x) (environment-ids e)))
    (environment-vals (list->vector (map val-transf (vector->list (environment-vals e)))))
    (environment-env  (if (environment-env e) 
                          (env-copy (environment-env e) val-transf)
                          #f))))

;; creates empty environment
(define (make-empty-env)
  (make-environment '() (list->vector '()) #f))

;; new environment with new mappings
(define new-env
  (lambda (ids vals)
    (extend-env ids vals (make-empty-env)))) 

;; extend environment with new mappings
(define extend-env
  (lambda (ids vals env)
    (if (eq? env (get-root-env))
        (new-env ids vals) ;; extending root, start new tree
        (make-environment ids (list->vector vals) env))))

;; imperative update of the given environment with a new mapping
(define augment-env
  (lambda (id val env)
    (set-environment-ids! env (cons id (environment-ids env)))
    (set-environment-vals! env (list->vector (cons val (vector->list (environment-vals env)))))))

;; lookup id in environment, apply action if found
(define env-lookup
  (lambda (action id env)
    (if env
        (let ((pos (find-position id (environment-ids env))))
          (if (number? pos)
              (action pos env) ;; read or write
              (env-lookup action id (environment-env env))))
        (root-env-lookup action id)))) ;; failed: look up in root

;; lookup id in root environment, error if not found
(define root-env-lookup
  (lambda (action id)
    (let* ((root (get-root-env))
           (pos (find-position id (environment-ids root))))
      (if (number? pos)
          (action pos root) ;; read or write
          (error 'env-lookup "No binding for ~s" id)))))
        
;; get var in given env
(define env-get
  (lambda (id env)
    (env-lookup (lambda (pos env) 
                  (vector-ref (environment-vals env) pos))
                id env)))

;; set var in given env
(define env-set!
  (lambda (id env nval)
    (env-lookup (lambda (pos env) 
                  (vector-set! (environment-vals env) pos nval))
                id env)))

;; get var in root env
(define root-env-get
  (lambda (id)
    (env-get id (get-root-env))))

;; add global variable (binding in root env) [not used]
;(define add-global-var
;  (lambda (id val)
;    (augment-env id val (get-root-env))))


(define (print-env env)
  (if env 
      (printf "~a" (environment-ids env))
      (print-env (environment-env env))))
  

;; HELPER FUNCTIONS

(define find-if
  (lambda (pred lst)
    (cond ((null? lst)      #f)
	  ((pred (car lst)) (car lst))
	  (else             (find-if pred (cdr lst))))))

(define collect-if
  (lambda (pred lst)
    (cond ((null? lst)      '())
	  ((pred (car lst)) (cons (car lst) (collect-if pred (cdr lst))))
	  (else             (collect-if pred (cdr lst))))))

(define find-position
  (lambda (sym lst)
    (let loop ((tail lst) (pos 0))
      (cond ((null? tail) #f)
	    ((eqv? sym (car tail)) pos)
	    (else (loop (cdr tail) (+ pos 1)))))))

(define find-last-position
  (lambda (sym lst)
    (let loop
	((tail lst) (curpos 0) (lastpos #f))
      (cond
       ((null? tail) lastpos)
       ((eqv? sym (car tail))
	(loop (cdr tail) (+ curpos 1) curpos))
       (else (loop (cdr tail) (+ curpos 1) lastpos))))))

