;; Servers are the representation of a host on the network
;; Each server has
;; - its own console window
;; - a root environment
;; - a list of host properties
;; - a list of exported functions
;; - one thread for dequeuing incoming function calls
;; - one thread per function call being executed
;;   (all calls to current-thread below -- except in start-server-init -- refer to this thread)

;associaton list thread value -> environment
(define root-env-store '())

(define (register-root-env key env)
  (set! root-env-store (cons (cons key env) root-env-store)))

(define (get-root-env)
  (let ((result (assoc (lookup-server-thread (current-thread)) root-env-store)))
    (if (eq? result #f)
        (error 'should-not-happen)
        (cdr result))))                


;; hosts are represented by their properties, which are association lists
;; with 2 default mapping 'host and 'port
(define (new-host name port)
  `((name . ,name) (port . ,port)))

(define (host-name h)
  (cdr (assq 'name h)))

(define (host-port h)
  (cdr (assq 'port h)))

;associaton list thread value -> host properties
(define props-store '())

(define (register-host-properties key props)
  (set! props-store (cons (cons key props) props-store)))

(define (get-current-host)
  (get-host-properties (current-thread)))

(define (get-host-properties key)
  (let ((result (assoc (lookup-server-thread key) props-store)))
    (if (eq? result #f)
        (error 'get-host-properties "No such server ~a" key)
        (cdr result))))
  

;; global list of hosts -- inspired by rmiregistry
;; maps host ids (host:port) to server thread ids.
;; not truly distributed as it assumes all server threads run within one vm
;; implementing this as a true remote server is straightforward
(define global-host-list '())

;; register a host id (eg. "localhost:8080") with a key (the thread id of that server)
;; the key can then be used to lookup eg. in the exported function lists of that host.
(define (register-host-global id key)
  (set! global-host-list (cons (cons id key) global-host-list)))

;; given a host id (eg. "localhost:8080") returns the id of that host thread
(define (lookup-host id)
    (let ((result (assoc id global-host-list)))
      (if (eq? result #f)
          (error 'lookup-host "unknown host ~a" id)
          (cdr result))))

;associaton list thread value -> global exported function list
(define global-exported-functions-list '())

(define global-exported-function-sema (make-semaphore 1))

;; thread-safe adding to global-exported-functions-list
(define (create-exported-functions-list key)
  (semaphore-wait global-exported-function-sema)
  (set! global-exported-functions-list (cons (cons key (vector '())) global-exported-functions-list))
  (semaphore-post global-exported-function-sema))

;; thread-safe only for lookup into global-exported-functions-list
(define (get-exported-functions-list-of key)
  (semaphore-wait global-exported-function-sema)
  (let ((binding (assoc (lookup-server-thread key) global-exported-functions-list)))
    (semaphore-post global-exported-function-sema)
    (if (eq? binding #f)
        (error 'get-exported-functions-list "No such server ~a" key)
        (vector-ref (cdr binding) 0))))                

(define (get-exported-functions-list)
  (get-exported-functions-list-of (current-thread)))

;; thread-safe only for lookup into global-exported-functions-list
(define (set-exported-functions-list! new-list)
  (semaphore-wait global-exported-function-sema)
  (let ((binding (assoc (lookup-server-thread (current-thread)) global-exported-functions-list)))
    (semaphore-post global-exported-function-sema)
    (if (eq? binding #f)
        (error 'set-exported-functions-list! "Setting exported functions in thread ~s" (current-thread))
        (vector-set! (cdr binding) 0 new-list))))

(define exported-function-sema (make-semaphore 1))

;; thread-safe way to add an exported function (current host)
(define (add-exported-function function stub)
  (semaphore-wait exported-function-sema)
  (let ((name (fun-stub-name stub)))
    (set-exported-functions-list! (cons (cons name function) (get-exported-functions-list))))
  (semaphore-post exported-function-sema))


;; lookup exported function on this host -- return the actual closure
(define (lookup-exported-function name)
  (lookup-exported-function-with name get-exported-functions-list))

;; lookup exported function on the given host -- return stub
(define (lookup-function-on-host host port name)
  (let* ((id  (string-append host ":" port))
         (key (lookup-host id)))
    (let* ((fun   (lookup-exported-function-with  ;; will fail if host does not provide the function
                   name 
                   (lambda () (get-exported-functions-list-of key))))
           (props (get-host-properties key))
           (argcount (arg-count fun)))
      (make-fun-stub props name argcount))))

;; thread-safe lookup of an exported function, by its name
;; the second parameter is the function that obtains the list
(define (lookup-exported-function-with name obtain-list)
  (semaphore-wait exported-function-sema)
  (let ((binding (assoc name (obtain-list)))) 
    (semaphore-post exported-function-sema)
    (if (eq? binding #f)
        (error 'lookup-exported-function "No such exported function ~s" name))
    (cdr binding)))

;; creates a stub for the given closure, exported locally using the given name
(define (export-named-function f name)
  (let ((stub (make-fun-stub (get-current-host) name (arg-count f))))
    (begin (add-exported-function f stub)
           (server-printf "Exporting named function ~a~n" name)
           stub)))

;; creates a stub for the given closure, exported locally using a generated name 
;; if the given function is in fact already a stub, just return the stub
(define (export-function f)
  (if (fun-stub? f) f
      (export-named-function f (generate-new-func-name))))

(define func-name-counter 0)
(define func-name-counter-sema (make-semaphore 1))
(define (generate-new-func-name)
  (begin
    (semaphore-wait func-name-counter-sema)
    (set! func-name-counter (+ func-name-counter 1))
    (semaphore-post func-name-counter-sema)
    (string-append "exp_fun_" (number->string func-name-counter))))


;root-env and exported-functions-list are kept per server thread 
;but each server spawns a handler thread per call, so we need an assocation list.
(define handler-thread-registry '())

(define handler-thread-sema (make-semaphore 1))

(define (register-handler-thread child parent)
  (semaphore-wait handler-thread-sema)
  (set! handler-thread-registry (cons (cons child parent) handler-thread-registry))
  (semaphore-post handler-thread-sema))

(define (unregister-handler-thread child)
  (semaphore-wait handler-thread-sema)
  (set! handler-thread-registry (filter (lambda (el) (not (equal? (car el) child))) handler-thread-registry))
  (semaphore-post handler-thread-sema)) 

(define (lookup-server-thread handler-thread)
    (let ((binding (assoc handler-thread handler-thread-registry)))
    (if (eq? binding #f)
        (error 'lookup-server-thread "Failed lookup of exported server thread for ~s.~n You are probably interpreting something in the console instead of on a server." handler-thread)
        (cdr binding))))                
 
;; thread-safe server->console association list
(define server-console-list '())

;; required to avoid race conditions in the user interface library when doing (send console-text insert foo)
(define server-print-sema (make-semaphore 1))

(define (add-server-console id console)
  (semaphore-wait server-print-sema)
  (set! server-console-list (cons (cons id console) server-console-list))
  (semaphore-post server-print-sema)
  )

(define (open-userinterface host port server-id)
  (let* ((identifier (string-append host ":" port))
         (frame (new frame% [label (string-append "Console for " identifier)] [width 800] [height 400]))
         (canvas (new editor-canvas% [parent frame]))
         (console-text (new text%))
         (menu-bar (new menu-bar% [parent frame]))
         (menu-edit (new menu% [label "Edit"] [parent menu-bar])))
    (send canvas set-editor console-text)
    (append-editor-operation-menu-items menu-edit #t)
    (send frame show #t)
    (send console-text insert "Server startup...")
    (send console-text insert #\newline)
    (add-server-console server-id console-text)))

(define server-printf
  (lambda args
    (semaphore-wait server-print-sema)
    (let ((binding (assoc (lookup-server-thread (current-thread)) server-console-list)))
      (semaphore-post server-print-sema)
      (if (eq? binding #f)
          (error 'server-printf "No console for ~s" (current-thread))
          (let ((editor (cdr binding)))
            (semaphore-wait server-print-sema)
            (send editor begin-edit-sequence)
            (send editor insert (apply format args))
            (send editor end-edit-sequence)
            (semaphore-post server-print-sema)
            )))))

;start-server is thread-safe 
(define (start-server host port)
  (start-server-init host port #f))
  
(define (start-server-init host port code)
  (let* ((creation-sema (make-semaphore))
         (server-startup-sema (make-semaphore))
         (queue (make-comms-queue host port))
         (props (new-host host port))
         (server-id (thread (lambda ()
                              (letrec ((loop (lambda () (begin (accept-call queue) (loop)))))
                                (register-handler-thread (current-thread) (current-thread))
                                (register-host-properties (current-thread) props)
                                (register-host-global (string-append host ":" port) (current-thread))
                                (semaphore-wait creation-sema)
                                (server-printf "Server adding host and port~n")
                                (run-init `(begin 
                                             (define current-host (quote ,props))
                                             (export-as (fun (thunk) (thunk)) "eval")
                                             )
                                          code) 
                                (semaphore-post server-startup-sema)
                                (server-printf "Server startup complete~n")
                                (loop))))))
    (open-userinterface host port server-id)
    (create-exported-functions-list server-id)
    (semaphore-post creation-sema)
    (semaphore-wait server-startup-sema)
    ))
